import numba
import numpy as np


EPSILON = 1e-6

@numba.jit(nopython=True)
def distance_trapz(prob1, prob2, times):
    """
    Calculates the area between two curves estimated by the trapezoidal rule
    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : 1d array
        The times at which the curves are evaluated

    Returns
    -------
    area : float
        The area between the curves
    """
    area = np.trapz(np.abs(prob1 - prob2), times)
    return area/((times[-1] - times[0]) + EPSILON)

@numba.jit(nopython=True)
def distance_euler(prob1, prob2, times):
    """
    Calculates the area between two curves estimated by the euler rule
    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : 1d array
        The times at which the curves are evaluated

    Returns
    -------
    area : float
        The area between the curves
    """

    area = np.sum((np.abs(prob1 - prob2)) * np.diff(add_zero(times)))
    return area/((times[-1] - times[0]) + EPSILON)

@numba.jit(nopython=True)
def add_zero(times):
    """
    This function concatenates 0 to the times array
    :param times:
    :return: return the times array with 0 concatenated
    """
    return np.concatenate((np.array([0]),times))



@numba.jit(nopython=True)
def distance_euclidean(prob1, prob2, times = None):
    """
    Calculates the euclidean distance between two curves at respective timepoints

    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : None
        times is not needed for this distance

    Returns
    -------
    euclidean distance : float
        The euclidean distance between the curves
    """
    print("Using Euler distance")
    return np.sqrt(np.sum((prob1 - prob2)**2))

@numba.jit(nopython=True)
def distance_cross_entropy(prob1, prob2, times = None):
    """
    Calculates the cross entropy between two curves at respective timepoints

    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : None
        times is not needed for this distance

    Returns
    -------
    cross entropy : float
        The cross entropy between the curves

    """
    return -np.sum(prob1*np.log(prob2 + EPSILON))


@numba.jit(nopython=True)
def distance_cross_entropy_from_mean(prob1, prob2, times = None):
    """
    The cross entropy between two curves and the mean of the two curves

    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : None
        times is not needed for this distance

    Returns
    -------
    cross entropy : float
        The cross entropy between the curves and the mean of the curves
    """
    mean_prob = (prob1 + prob2)/2
    return distance_cross_entropy(prob1, mean_prob) + distance_cross_entropy(prob2, mean_prob)


@numba.jit(nopython=True)
def distance_kl(prob1, prob2 , times= None):
    """
    KL divergence of second probability from first probability

    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : None
        times is not needed for this distance

    Returns
    -------
    KL divergence : float
        The KL divergence of the second probability from the first probability
    """
    return np.sum(prob1*np.log(prob1/(prob2 + EPSILON) + EPSILON))

@numba.jit(nopython=True)
def distance_shannon_jensen(prob1, prob2, times = None):
    """
    The symmetric version of KL divergence

    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : None
        times is not needed for this distance

    Returns
    -------
    KL divergence : float
        The symmetric version of KL divergence
    """
    mean_prob = (prob1 + prob2)/2
    return distance_kl(prob1, mean_prob,times = None) + distance_kl(prob2, mean_prob , times = None)

@numba.jit(nopython=True)
def distance_max(prob1, prob2, times = None):
    """
    The maximum difference between two curves

    Parameters
    ----------
    prob1 : 1d array
        The first curve
    prob2 : 1d array
        The second curve
    times : None
        times is not needed for this distance

    Returns
    -------
    max difference : float
        The maximum difference between the two curves

    """
    return np.max(np.abs(prob1 - prob2))





import pandas as pd
from pycox.evaluation import EvalSurv
from sklearn.base import BaseEstimator
from sklearn.linear_model import LogisticRegression
from sksurv.ensemble import RandomSurvivalForest
from sksurv.nonparametric import kaplan_meier_estimator
from sksurv.tree import SurvivalTree
from .indexcalculator import IndexCalculator
from proxsurv.utils import predict_probability_at_times
import numpy as np
from sklearn.model_selection import train_test_split
from sksurv.linear_model import CoxnetSurvivalAnalysis
from .npsurvival_models import KNNSurvival


class CobraSurvival(BaseEstimator):
    """
    This is COBRA implementation for survival analysis, it is a meta-learner
    that combines the predictions of several survival models to produce a
    final prediction. The algorithm is described in the paper:



    .. math::


        \\hat{S}_{COBRA}(t | x; \\cdot ) = \\prod_{t' \\in \\mathcal{Y}_{\\Gamma}(x; \\cdot)}
         \\left( 1 - \\frac{\\mathcal{D}_{\\Gamma}(t' | x; \\cdot )}{\mathcal{R}_{\\Gamma}(t' | x; \\cdot)} \\right)^{\\mathbb{I}(t' \\leq t)}




    See the :ref:`User Guide </user_guide/cobrasurvival.ipynb>`, and the research paper
    [1]_  for further description.

    Parameters
    ----------
    epsilon: float , default is 0.12
        The :math:`\\epsilon` proximity which checks if the two machines survival prediction are
        close :math:`d(s_1,s_2) < \\epsilon`

    alpha: integer , default is 4
        The :math:`\\alpha` the number of machines which are in consensus of :math:`\\epsilon` proximity

    machines: list of scikit-learn type of machines, default is "default"
        machines, parameter, set to default for now, The defaults are Random Survival Forest, Cox Ridge,
        Cox Lasso, Survival Tree, KNN Survival but user can provide their own machines with conditions that it have a
        predict_survival_function method which returns a numpy array of shape (n_samples,n_times), the survival
        prediction at times of unique times of training data.

    distance_function: string or callable, default is "euler"
        distance function, parameter, set to default for now, The default is the area type
        of norm, as used in the paper, but user can provide their own distance function which takes two numpy
        arrays of shape (n_times,) and returns a float. The distance function can be "trapz", "euclidean",
        "cross_entropy", "cross_entropy_from_mean", "shannon_jensen", "kl", "max", "euler" or a callable function.


    Attributes
    ----------
    unique_sorted_times_l: 1d array
        Unique sorted time points of happening of an event or censoring in dataset :math:`D_l`.
    unique_sorted_times_k: 1d array
        Unique sorted time points of happening of an event or censoring in dataset :math:`D_k`.
    pred_l : 3d array
        The survival prediction of machines on dataset :math:`D_l` of shape (n_samples,n_estimators,n_times).
    epsilon : float
        The :math:`\\epsilon` proximity which checks if the two machines survival prediction sre close :math:`d(s_1,s_2) < \\epsilon`.
    alpha : integer
        The :math:`\\alpha` the number of machines which are in consensus of $\\epsilon$ proximity.
    n_estimators : int , default is 5
        The number of machines.
    indexer : A IndexCalculator object
        The IndexCalculator object which is used to calculate the index.
    train_time : 1d array
        The  time points of dataset :math:`D_l`.
    train_event : 1d array
        A boolean indicator for censoring True represents the happening of an event of dataset :math:`D_l`.



    See also
    --------
    cbrsurv.distance_functions : The distance functions used in COBRA
        A single survival tree.

    Notes
    -----
    The COBRA algorithm is described in the paper [1]_. We must know that as we increase :math:`\\epsilon` the
    survival curve will become more like the population survival curve.


    References
    ----------
    .. [1] Rahul Goswami, and Arabin K. Dey. "Area-norm COBRA on Conditional Survival Prediction."
        ArXiv, (2023). Accessed September 6, 2023. https://arxiv.org/abs/2309.00417.


    """

    def __init__(self, epsilon=0.12, alpha=4, machines="default", distance_function="euler"):

        self.epsilon = epsilon
        self.alpha = alpha
        self.machines = machines
        self.distance_function = distance_function

    def fit(self, X, y, n_quantiles=10, l_by_n=0.5, experimental=False):
        """
        This part does the splitting the dataset in two parts $X_l$ and $X_k$ and $y_l$ and $y_k$,
        the splitting is done strategically to ensure that censoring and event times are distributed
        equally in both the splits, Secondly this function train the initial models on $D_k$ dataset



        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
        y : structured numpy array of shape (n_samples,)
            The target variable with the first field as the event indicator and the second field as the event time
        n_quantiles : int, default=10
            The number of quantiles to be used for splitting the data, defult is set to 10
        l_by_n : float, default=0.5
            The split ratio, defult is set to 0.5
        experimental : bool, default=False
            If set to True the models are trained on whole dataset and no splitting will happen , default is False

        Returns
        -------
        self : object

        """



        if not isinstance(n_quantiles, int) or n_quantiles <= 1:
            raise ValueError("The number of quantiles must be an integer greater than 1.")

        # Ensure that the split ratio is within the valid range of 0 to 1
        if not (0 < l_by_n < 1):
            raise ValueError("The split ratio must be a value between 0 and 1 (exclusive).")

        # Ensure that the input data 'X' is a numpy array
        if not isinstance(X, np.ndarray) and X.ndim != 2:
            raise TypeError("Input 'X' must be a numpy array.")

        # Ensure that the target variable 'y' is a numpy array
        if not isinstance(X, np.ndarray):
            raise TypeError("Target variable 'y' must be a numpy array.")

        assert X.shape[0] == y.shape[
            0], "The number of samples in 'X' and 'y' must be equal, but got {0} and {1} respectively.".format(
            X.shape[0], y.shape[0])

        _, self.n_features_in_ = X.shape

        while True:
            try:
                time = y[y.dtype.names[1]]
                strat_array = pd.qcut(time, q=n_quantiles, labels=False, duplicates='drop')
                result = [str(x) + str(y) for x, y in zip(strat_array, (y[y.dtype.names[0]]))]
                Xl, Xk, yl, yk = train_test_split(X, y, test_size=l_by_n, stratify=result, random_state=2)
                break
            except:
                print("Warning : The number of quantiles is too high, reducing the number of quantiles by 1")
                n_quantiles = n_quantiles - 1

        if experimental:
            Xl, Xk = X, X
            yl, yk = y, y

        self.Xl = Xl  # needed for covariate relevance calculation

        self.train_time = yl[yl.dtype.names[1]]
        self.train_event = yl[yl.dtype.names[0]].astype(bool)
        self.unique_sorted_times_l = np.sort(np.unique(yl[yl.dtype.names[1]])).astype(float)
        self.unique_sorted_times_k = np.sort(np.unique(yk[yk.dtype.names[1]])).astype(float)
        self._fit_initial_machines(Xk, yk)
        self.pred_l = self._get_initial_predictions(Xl)

        return self

    def _fit_initial_machines(self, Xk, yk):
        """
        Fit the initial machines on the dataset $D_k$

        Parameters
        ----------
        Xk : array-like of shape (n_samples, n_features)
            The input samples.
        yk : structured numpy array of shape (n_samples,)
            The target variable with the first field as the event indicator and the second field as the event time
        """
        if self.machines == "default":
            self.rf = RandomSurvivalForest(n_estimators=100, random_state=42)
            self.rf.fit(Xk, yk)

            self.knn = KNNSurvival()
            self.knn.fit(Xk, np.array([yk[yk.dtype.names[1]], yk[yk.dtype.names[0]]]).T)

            self.coxridge = CoxnetSurvivalAnalysis(l1_ratio=0.0000001, fit_baseline_model=True)
            self.coxridge.fit(Xk, yk)

            self.coxlasso = CoxnetSurvivalAnalysis(l1_ratio=1.0, fit_baseline_model=True)
            self.coxlasso.fit(Xk, yk)

            self.tree = SurvivalTree(random_state=42)
            self.tree.fit(Xk, yk)
        else:
            if isinstance(self.machines, list):
                for machine in self.machines:
                    assert callable(machine.fit) and callable(machine.predict_survival_function), \
                        f"The machine must have fit and predict_survival_function method , but for {machine} it is not the case"
                    assert hasattr(machine, "unique_times_"), \
                        f"The machine must have unique_times_ attribute , but for {machine} it is not the case"
                self.machines = self.machines
            else:
                raise TypeError(
                    "The machines must be a list of machines , which have fit and predict_survival_function method")
            fitted_machines = []
            for machine in self.machines:
                try:
                    machine.fit(Xk, yk)
                    fitted_machines.append(machine)
                except:
                    print(f"Warning : The machine {machine} is not fitted on the dataset")
            self.machines = fitted_machines

    def _get_initial_predictions(self, X):
        """
        Get the initial predictions from the machines
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
        """
        if self.machines == "default":
            self.rf_pred = self.rf.predict_survival_function(X, return_array=True)
            self.knn_pred = self.knn.predict_surv(X, self.unique_sorted_times_k)
            self.tree_pred = self.tree.predict_survival_function(X, return_array=True)
            self.ridge_pred = self.coxridge.predict_survival_function(X, return_array=True)
            self.lasso_pred = self.coxlasso.predict_survival_function(X, return_array=True)
            self.n_estimators = 5
            result = np.array([self.rf_pred, self.knn_pred, self.tree_pred, self.lasso_pred, self.ridge_pred])
        else:
            pred_list = []
            for machine in self.machines:
                try:
                    pred = machine.predict_survival_function(X)
                    pred = predict_probability_at_times(machine.unique_times_, pred, self.unique_sorted_times_k)
                    pred_list.append(pred)
                except:
                    print(f"Warning : The machine {machine} is not fitted on the dataset")

            result = np.array(pred_list)

            self.n_estimators = len(pred_list)
            del pred_list

        return np.transpose(result, (1, 0, 2)).astype(np.float32)

    def predict(self, X):
        """
        Predict the survival function for a given set of covariates

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
        """

        query_pred = self._get_initial_predictions(X)

        self.indexer = IndexCalculator(query_pred, self.pred_l, self.unique_sorted_times_k)
        index_arr = self.indexer.get_index(self.epsilon, self.alpha, self.distance_function)

        return self._get_survival_function(index_arr)

    def _get_survival_function(self, index_arr):
        """
        Get the survival function from the selected machines

        Parameters
        ----------
        index_arr : array-like of shape (n_samples, n_samples in $D_l$)
            The index array which is used to select the proximity observations from :math:`D_l`
        """
        estimate = np.zeros((index_arr.shape[0], len(self.unique_sorted_times_l)))
        for i in range(index_arr.shape[0]):
            time, prob = kaplan_meier_estimator(self.train_event[index_arr[i]], self.train_time[index_arr[i]])
            estimate[i] = predict_probability_at_times(time, prob, self.unique_sorted_times_l)
        return estimate

    def get_covariate_relevance(self, X=None):
        """
        Get the covariate relevance of the covariates
        This function computes the relevance with respect to given covariates, if no covariates are given then it
        computes the relevance with respect to average of all the in the set $D_l$

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        """
        if X is None:
            query_x = self.Xl.mean(axis=0)
            query_pred = self._get_initial_predictions(query_x.reshape(1, -1))
        else:
            query_pred = self._get_initial_predictions(X.reshape(1, -1))
        index_arr = IndexCalculator(query_pred, self.pred_l, self.unique_sorted_times_k).get_index(self.epsilon,
                                                                                                   self.alpha,
                                                                                                   self.distance_function)
        model = LogisticRegression(fit_intercept=False)
        model.fit(self.Xl, index_arr)
        return model.coef_

    def _get_ev(self, X, y):
        """
        The package uses EvalSurv package from pycox to calculate the concordance and integrated brier score, and various
        other metrics that are available in the package, also it uses ploting of pycox to plot the survival function

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
        y : structured numpy array of shape (n_samples,)
            The target variable with the first field as the event indicator and the second field as the event time
        """

        prediction = self.predict(X)
        prediction = pd.DataFrame(prediction.T, index=self.unique_sorted_times_l)
        ev = EvalSurv(prediction, y[y.dtype.names[1]], y[y.dtype.names[0]], censor_surv='km')
        return ev

    def score(self, X, y, type="ibs"):
        """
        The score function provides the integrated brier score by default, if type is set to concordance then it provides
        the concordance index

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
        y : structured numpy array of shape (n_samples,)
            The target variable with the first field as the event indicator and the second field as the event time
        type : string, default is "ibs"
            The type of score to be calculated, it can be "ibs" or "concordance"

        Returns
        -------
        score : float
            The score of the model
        """
        ev = self._get_ev(X, y)
        if type == "ibs":
            return ev.integrated_brier_score(self.unique_sorted_times_l)
        elif type == "concordance":
            return ev.concordance_td()

    def plot(self, X, index):
        """
        This plot the survival function of the given covariates

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
        index : integer
            The index of the covariate to be plotted
        """
        event = np.random.random(X.shape[0]) < 0.5
        time = np.random.randint(0, 365, size=X.shape[0])
        y = np.array(list(zip(event, time)), dtype=[('event', '?'), ('time', '<f8')])
        ev = self._get_ev(X, y)
        ev[index].plot_surv()






import numba
import numpy as np
from numba import prange


@numba.jit(nopython=True)
def index_calculator_helper(i, pred_l,test_pred ,n_estimators, alpha, epsilon, times , distance_func):
    """
    A helper function for index_calculator, it is used for canculating the index of one row, i.e. one test data
    proximity to all training data



    Parameters
    ----------
    i : int
        The observation index
    pred_l : 3d array
        Predictions of all training data
    test_pred : 3d array
        Predictions of all test data
    n_estimators : int
        number of estimators
    alpha : int
        The number of base estimators to be in consensus
    epsilon : float
        the proximity threshold
    times : 1d array
        The times at which the predictions are made
    distance_func : function
        The distance function

    Returns
    -------
    index_arr_row : 1d array
        The index of the test data to all observation in :math:`D_l` which is in :math:`\epsilon,\alpha)` proximity
    """
    index_arr_row = np.zeros(pred_l.shape[0], dtype=np.bool_)

    flag = False

    for j in prange(pred_l.shape[0]):
        n_estimators_consensus = 0

        for k in range(n_estimators):
            if distance_func(test_pred[i, k, :], pred_l[j, k, :], times) <= epsilon:
                n_estimators_consensus += 1

        if n_estimators_consensus >= alpha:
            index_arr_row[j] = True
            flag = True

    if not flag:
        index_arr_row[:] = True

    return index_arr_row


@numba.jit(nopython=True)
def _predict_probability_at_times(uniq_times, probability, times):
    """
    A helper function for predict_probability_at_times, it is used for
    one dimensional array



    Parameters
    ----------
    uniq_times : 1d array
        times at which probability is given
    probability : 1d array
        probability at uniq_times
    times : 1d array
        compute probability at times
    """
    below_range = times < uniq_times[0]
    above_range = times > uniq_times[-1]
    prob_new = np.empty(len(times), dtype=float)
    prob_new[below_range] = 1.0
    slope = (probability[-1] - 1.0 + 1e-16) / (uniq_times[-1] - 0.0 + 1e-16)
    intercept = 1.0 - slope * 0.0
    # times_original = times
    prob_new[above_range] = slope * times[above_range] + intercept
    times = times[~below_range & ~above_range]
    idx = np.searchsorted(uniq_times, times)
    eps = np.finfo(uniq_times.dtype).eps
    exact = np.absolute(uniq_times[idx] - times) < eps
    idx[~exact] -= 1
    prob_new[~below_range & ~above_range] = probability[idx]
    prob_new[prob_new < 0] = 0
    return prob_new


def predict_probability_at_times(uniq_times, probability, times):
    """
    Predict the probability at times

    Parameters
    ----------
    uniq_times : 1d array
        times at which probability is given
    probability : 1d or 2d array
        probability at uniq_times
    times : 1d array
        compute probability at times
    """
    if len(probability.shape) == 1:
        return _predict_probability_at_times(uniq_times, probability, times)
    else:
        prob_new = np.zeros((probability.shape[0], times.shape[0]))
        for i in range(probability.shape[0]):
            prob_new[i] = _predict_probability_at_times(uniq_times,
                                                        probability[i], times)
        return prob_new





# -*- coding: utf-8 -*-

"""pr0xsurv - A package for proximal survival analysis"""

__author__ = """Rahul Goswami"""
__email__ = 'yuvrajiro@gmail.com'
__version__ = '0.0.3'

print(" We are to sorry to inform you we have not yet implemented the package.")


import numpy as np
from joblib import Parallel, delayed

from proxsurv.cbrsurv.distance_functions import distance_euclidean, distance_cross_entropy, \
    distance_cross_entropy_from_mean, distance_shannon_jensen, distance_kl, distance_max, distance_euler, distance_trapz
from ..utils import index_calculator_helper


class IndexCalculator(object):
    """Class for calculating the index of a given CBR object.

    The index is calculated by checking whether the $(\epsilon , \alpha)$
    proximity is satisfied for each case in the CBR.


    Attributes:
        epsilon: The epsilon value to use for the index calculation.
        alpha: The alpha value to use for the index calculation.
        distance: The distance metric to use for the index calculation.
    """
    def __init__(self,query_pred,pred_l,times):
        """Initializes the IndexCalculator with the given CBR object.

        Args:
            epsilon: The epsilon value to use for the index calculation.
            alpha: The alpha value to use for the index calculation.
            distance: The distance metric to use for the index calculation.It
            is set to "area" by default, but a function can be passed instead.
            other options are "euclidean", "cross_entropy" , a symmetric version
            of KL divergence named "shannon_jensen" , and a logrank statistics.

        """
        self.times = times
        self.n_estimators = pred_l.shape[1]
        assert query_pred.shape[1] == pred_l.shape[1] and query_pred.shape[2] == pred_l.shape[2], \
            "query_pred and pred_l have different dimensions"
        assert query_pred.shape[2] == times.shape[0], \
            "query_pred and times have different dimensions"
        assert query_pred.shape[1] == self.n_estimators, \
            "query_pred and n_estimators have different dimensions"

        self.pred_l = pred_l
        self.query_pred = query_pred

        self.index_arr = np.ones((query_pred.shape[0], pred_l.shape[0]), dtype=np.bool_)



    def hypers(self,epsilon,alpha,distance = "area"):
        self.epsilon = epsilon
        self.alpha = alpha
        if isinstance(distance, str):
            if distance == "trapz":
                self.distance = distance_trapz
            elif distance == "euclidean":
                self.distance = distance_euclidean
            elif distance == "cross_entropy":
                self.distance = distance_cross_entropy
            elif distance == "cross_entropy_from_mean":
                self.distance = distance_cross_entropy_from_mean
            elif distance == "shannon_jensen":
                self.distance = distance_shannon_jensen
            elif distance == "kl":
                self.distance = distance_kl
            elif distance == "max":
                self.distance = distance_max
            elif distance == "euler":
                self.distance = distance_euler
            else:
                raise ValueError("distance is not a valid string")
        elif callable(distance):
            self.distance = distance
        else:
            raise ValueError("distance is not a valid function or string")


    def calculate_index(self):
        """
        For a given set of prediction, this function calculates whether  the prediction is in $(\epsilon,\alpha)$ proximity
            :param test_pred: A prediction by machines
            :return: A boolean array of shape (n_samples,n_samples in $D_l$)
        """

        index = Parallel(n_jobs=-1)(delayed(index_calculator_helper)(i, self.pred_l, self.query_pred, self.n_estimators, self.alpha, self.epsilon,
                                            self.times, self.distance) for i in range(self.query_pred.shape[0]))
        self.index_arr = np.array(index)
        return self.index_arr

    def get_index(self, epsilon , alpha , distance = "area"):
        self.hypers(epsilon,alpha,distance)
        return self.calculate_index()




"""
The credits for the following code, goes to the author of the following repository:
- https://github.com/georgehc/npsurvival/tree/master

We do not claim the authorship, and we have only modified the code to suit our needs.

The code is attached in package, because the original repository may change in future
and we want to keep the code intact.

"""






















import numpy as np
from sklearn.neighbors import NearestNeighbors


class BasicSurvival():
    def __init__(self):
        self.tree = None

    def fit(self, y):
        self.tree = _fit_leaf(y)

    def predict_surv(self, times, presorted_times=False,
                      limit_from_left=False):
        """
        Computes the Kaplan-Meier survival probability function estimate at
        user-specified times.

        Parameters
        ----------
        times : 1D numpy array
            Times to compute the survival probability function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the survival
            probability function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        Returns
        -------
        output : 1D numpy array
            Survival probability function evaluated at each of the times
            specified in `times`.
        """
        return _predict_leaf(self.tree, 'surv', times, presorted_times,
                             limit_from_left)

    def predict_cum_haz(self, times, presorted_times=False,
                        limit_from_left=False):
        """
        Computes the Nelson-Aalen cumulative hazard function estimate at
        user-specified times.

        Parameters
        ----------
        times : 1D numpy array
            Times to compute the cumulative hazard function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the
            cumulative hazard function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        Returns
        -------
        output : 1D numpy array
            Cumulative hazard function evaluated at each of the times
            specified in `times`.
        """
        return _predict_leaf(self.tree, 'cum_haz', times, presorted_times,
                             limit_from_left)


class KNNSurvival():
    def __init__(self, *args, **kwargs):
        """
        Arguments are the same as for `sklearn.neighbors.NearestNeighbors`.
        The simplest usage of this class is to use a single argument, which is
        `n_neighbors` for the number of nearest neighbors (Euclidean distance
        is assumed in this case). If you want to parallelize across different
        search queries, use the `n_jobs` keyword parameter (-1 to use all
        cores). To use other distances and for other details, please refer to
        the documentation for sklearn's `NearestNeighbors` class.

        *Important:* The prediction methods for this class use unweighted
        k-nearest neighbors, where "k" is set equal to the `n_neighbors`
        parameter.
        """
        self.NN_index_args = args
        self.NN_index_kwargs = kwargs
        self.NN_index = None

    def fit(self, X, y):
        """
        Constructs a nearest-neighbor index given training data (so that for
        a future data point, we can use the nearest-neighbor index to quickly
        find what the closest training data are to the future point).

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        y : 2D numpy array, shape = [n_samples, 2]
            Survival labels (first column is for observed times, second column
            is for event indicators). The i-th row corresponds to the i-th row
            in `X`.

        Returns
        -------
        None
        """
        self.train_y = y
        self.NN_index = NearestNeighbors(*self.NN_index_args,
                                         **self.NN_index_kwargs)
        self.NN_index.fit(X)

    def predict_surv(self, X, times, presorted_times=False,
                      limit_from_left=False, n_neighbors=None):
        """
        Computes the k-NN Kaplan-Meier survival probability function estimate
        at user-specified times.

        *Important:* The default number of nearest neighbors to use is whatever
        was specified in `args` or `kwargs` when creating an instance of this
        class (the "k" in k-NN)!

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        times : 1D numpy array
            Times to compute the survival probability function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the survival
            probability function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        n_neighbors : int, None, optional (default=None)
            Number of nearest neighbors to use. If set to None then the number
            used is whatever was passed into `args` or `kwargs` when creating
            an instance of this class.

        Returns
        -------
        output : 2D numpy array
            Survival probability function evaluated at each of the times
            specified in `times` for each feature vector. The i-th row
            corresponds to the i-th feature vector.
        """
        indices = self.NN_index.kneighbors(X, n_neighbors=n_neighbors,
                                           return_distance=False)
        train_y = self.train_y
        return np.array([_predict_leaf(_fit_leaf(train_y[idx]), 'surv', times,
                                       presorted_times, limit_from_left)
                         for idx in indices])

    def predict_cum_haz(self, X, times, presorted_times=False,
                        limit_from_left=False, n_neighbors=None):
        """
        Computes the k-NN Nelson-Aalen cumulative hazard function estimate at
        user-specified times.

        *Important:* The default number of nearest neighbors to use is whatever
        was specified in `args` or `kwargs` when creating an instance of this
        class (the "k" in k-NN)!

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        times : 1D numpy array
            Times to compute the cumulative hazard function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the
            cumulative hazard function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        n_neighbors : int, None, optional (default=None)
            Number of nearest neighbors to use. If set to None then the number
            used is whatever was passed into `args` or `kwargs` when creating
            an instance of this class.

        Returns
        -------
        output : 2D numpy array
            Cumulative hazard function evaluated at each of the times
            specified in `times` for each feature vector. The i-th row
            corresponds to the i-th feature vector.
        """
        indices = self.NN_index.kneighbors(X, n_neighbors=n_neighbors,
                                           return_distance=False)
        train_y = self.train_y
        return np.array([_predict_leaf(_fit_leaf(train_y[idx]), 'cum_haz',
                                       times, presorted_times, limit_from_left)
                         for idx in indices])


class KNNWeightedSurvival():
    def __init__(self, *args, **kwargs):
        """
        Arguments are the same as for `sklearn.neighbors.NearestNeighbors`.
        The simplest usage of this class is to use a single argument, which is
        `n_neighbors` for the number of nearest neighbors (Euclidean distance
        is assumed in this case). If you want to parallelize across different
        search queries, use the `n_jobs` keyword parameter (-1 to use all
        cores). To use other distances and for other details, please refer to
        the documentation for sklearn's `NearestNeighbors` class.

        *Important:* The prediction methods for this class use weighted
        k-nearest neighbors, where "k" is set equal to the `n_neighbors`
        parameter. The weights are specified through a kernel function K. In
        particular, the i-th nearest neighbor X_i for a test point x is given a
        weight of:
            K( (distance between x and X_i) / (distance between x and X_k) ).
        """
        self.NN_index_args = args
        self.NN_index_kwargs = kwargs
        self.NN_index = None

    def fit(self, X, y):
        """
        Constructs a nearest-neighbor index given training data (so that for
        a future data point, we can use the nearest-neighbor index to quickly
        find what the closest training data are to the future point).

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        y : 2D numpy array, shape = [n_samples, 2]
            Survival labels (first column is for observed times, second column
            is for event indicators). The i-th row corresponds to the i-th row
            in `X`.

        Returns
        -------
        None
        """
        self.train_y = y
        self.NN_index = NearestNeighbors(*self.NN_index_args,
                                         **self.NN_index_kwargs)
        self.NN_index.fit(X)

    def predict_surv(self, X, times, presorted_times=False,
                      limit_from_left=False, n_neighbors=None,
                      kernel_function=None):
        """
        Computes the weighted k-NN Kaplan-Meier survival probability function
        estimate at user-specified times.

        *Important:* The default number of nearest neighbors to use is whatever
        was specified in `args` or `kwargs` when creating an instance of this
        class (the "k" in k-NN)!

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        times : 1D numpy array
            Times to compute the survival probability function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the survival
            probability function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        n_neighbors : int, None, optional (default=None)
            Number of nearest neighbors to use. If set to None then the number
            used is whatever was passed into `args` or `kwargs` when creating
            an instance of this class.

        kernel_function : function, None, optional (default=None)
            Kernel function to use. None corresponds to unweighted k-NN
            survival analysis. If a function is specified, then the weighting
            function used is of the form
            "kernel(distance / distance to k-th nearest neighbor)".

        Returns
        -------
        output : 2D numpy array
            Survival probability function evaluated at each of the times
            specified in `times` for each feature vector. The i-th row
            corresponds to the i-th feature vector.
        """
        if kernel_function is None:
            kernel_function = lambda s: 1
        dists, indices = self.NN_index.kneighbors(X, n_neighbors=n_neighbors,
                                                  return_distance=True)
        train_y = self.train_y
        output = []
        n_times = len(times)
        for dist, idx in zip(dists, indices):
            max_dist = np.max(dist)
            weights = np.array([kernel_function(d / max_dist) for d in dist])
            zero_weight = (weights == 0)
            if zero_weight.sum() > 0:
                weights_subset = weights[~zero_weight]
                if weights_subset.size > 0:
                    output.append(
                        _predict_leaf(
                            _fit_leaf_weighted(train_y[idx[~zero_weight]],
                                               weights_subset),
                            'surv', times, presorted_times, limit_from_left))
                else:
                    output.append(np.ones(n_times))
            else:
                output.append(
                    _predict_leaf(
                        _fit_leaf_weighted(train_y[idx],
                                           weights),
                        'surv', times, presorted_times, limit_from_left))
        return np.array(output)

    def predict_cum_haz(self, X, times, presorted_times=False,
                        limit_from_left=False, n_neighbors=None,
                        kernel_function=None):
        """
        Computes the weighted k-NN Nelson-Aalen cumulative hazard function
        estimate at user-specified times.

        *Important:* The default number of nearest neighbors to use is whatever
        was specified in `args` or `kwargs` when creating an instance of this
        class (the "k" in k-NN)!

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        times : 1D numpy array
            Times to compute the cumulative hazard function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the
            cumulative hazard function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        n_neighbors : int, None, optional (default=None)
            Number of nearest neighbors to use. If set to None then the number
            used is whatever was passed into `args` or `kwargs` when creating
            an instance of this class.

        kernel_function : function, None, optional (default=None)
            Kernel function to use. None corresponds to unweighted k-NN
            survival analysis. If a function is specified, then the weighting
            function used is of the form
            "kernel(distance / distance to k-th nearest neighbor)".

        Returns
        -------
        output : 2D numpy array
            Cumulative hazard function evaluated at each of the times
            specified in `times` for each feature vector. The i-th row
            corresponds to the i-th feature vector.
        """
        if kernel_function is None:
            kernel_function = lambda s: 1
        dists, indices = self.NN_index.kneighbors(X, n_neighbors=n_neighbors,
                                                  return_distance=True)
        train_y = self.train_y
        output = []
        n_times = len(times)
        for dist, idx in zip(dists, indices):
            max_dist = np.max(dist)
            weights = np.array([kernel_function(d / max_dist) for d in dist])
            zero_weight = (weights == 0)
            if zero_weight.sum() > 0:
                weights_subset = weights[~zero_weight]
                if weights_subset.size > 0:
                    output.append(
                        _predict_leaf(
                            _fit_leaf_weighted(train_y[idx[~zero_weight]],
                                               weights_subset),
                            'cum_haz', times, presorted_times, limit_from_left))
                else:
                    output.append(np.zeros(n_times))
            else:
                output.append(
                    _predict_leaf(
                        _fit_leaf_weighted(train_y[idx],
                                           weights),
                        'cum_haz', times, presorted_times, limit_from_left))
        return np.array(output)


def _fit_leaf_weighted(y, weights):
    """
    Computes leaf node information given survival labels (observed times and
    event indicators) that have weights. This is for computing kernel variants
    of the Kaplan-Meier and Nelson-Aalen estimators.

    Parameters
    ----------
    y : 2D numpy array, shape=[n_samples, 2]
        The two columns correspond to observed times and event indicators.

    weights : 1D numpy array, shape=[n_samples]
        Nonnegative weights; i-th weight corresponds to the i-th row in `y`.

    Returns
    -------
    tree : dictionary
        The leaf node information stored as a dictionary. Specifically, the
        key-value pairs of this dictionary are as follows:
        - 'times': stores the sorted unique observed times
        - 'event_counts': in the same order as `times`, the number of events
            at each unique observed time
        - 'at_risk_counts': in the same order as `times`, the number of
            subjects at risk at each unique observed time
        - 'surv': in the same order as `times`, the Kaplan-Meier survival
            probability estimate at each unique observed time
        - 'cum_haz': in the same order as `times`, the Nelson-Aalen cumulative
            hazard estimate at each unique observed time
    """
    if y.size == 0:
        return {'times': sorted_unique_observed_times,
                'event_counts': event_counts,
                'at_risk_counts': at_risk_counts,
                'surv': surv_func,
                'cum_haz': cum_haz_func}

    if len(y.shape) == 1:
        y = y.reshape(1, -1)

    sorted_unique_observed_times = np.sort(np.unique(y[:, 0]))
    num_unique_observed_times = len(sorted_unique_observed_times)
    time_to_idx = {time: idx
                   for idx, time in enumerate(sorted_unique_observed_times)}
    event_counts = np.zeros(num_unique_observed_times)
    dropout_counts = np.zeros(num_unique_observed_times)
    at_risk_counts = np.zeros(num_unique_observed_times)
    at_risk_counts[0] = np.sum(weights)

    for (observed_time, event_ind), weight in zip(y, weights):
        idx = time_to_idx[observed_time]
        if event_ind:
            event_counts[idx] += weight
        dropout_counts[idx] += weight

    for idx in range(num_unique_observed_times - 1):
        at_risk_counts[idx + 1] = at_risk_counts[idx] - dropout_counts[idx]

    surv_prob = 1.
    cum_haz = 0.
    surv_func = np.zeros(num_unique_observed_times)
    cum_haz_func = np.zeros(num_unique_observed_times)
    for idx in range(num_unique_observed_times):
        frac = event_counts[idx] / at_risk_counts[idx]
        surv_prob *= 1 - frac
        cum_haz += frac
        surv_func[idx] = surv_prob
        cum_haz_func[idx] = cum_haz

    return {'times': sorted_unique_observed_times,
            'event_counts': event_counts,
            'at_risk_counts': at_risk_counts,
            'surv': surv_func,
            'cum_haz': cum_haz_func}


def _predict_leaf(tree, mode, times, presorted_times, limit_from_left=False):
    """
    Computes either the Kaplan-Meier survival function estimate or the
    Nelson-Aalen cumulative hazard function estimate at user-specified times
    using survival label data in a leaf node.

    Parameters
    ----------
    tree : dictionary
        Leaf node of a decision tree where we pull survival label information
        from.

    mode : string
        Either 'surv' for survival probabilities or 'cum_haz' for cumulative
        hazard function.

    times : 1D numpy array
        Times to compute the survival probability or cumulative hazard function
        at.

    presorted_times : boolean
        Flag for whether `times` is already sorted.

    limit_from_left : boolean, optional (default=False)
        Flag for whether to output the function evaluated at a time just to the
        left, i.e., instead of outputting f(t) where f is either the survival
        probability or cumulative hazard function estimate, output:
            f(t-) := limit as t' approaches t from the left of f(t').

    Returns
    -------
    output : 1D numpy array
        Survival probability or cumulative hazard function evaluated at each of
        the times specified in `times`.
    """
    unique_observed_times = tree['times']
    surv_func = tree[mode]

    if presorted_times:
        sort_indices = range(len(times))
    else:
        sort_indices = np.argsort(times)

    num_leaf_times = len(unique_observed_times)
    leaf_time_idx = 0
    last_seen_surv_prob = 1.
    output = np.zeros(len(times))
    if limit_from_left:
        for sort_idx in sort_indices:
            time = times[sort_idx]
            while leaf_time_idx < num_leaf_times:
                if unique_observed_times[leaf_time_idx] <= time:
                    last_seen_surv_prob = surv_func[leaf_time_idx]
                    leaf_time_idx += 1
                else:
                    break
            output[sort_idx] = last_seen_surv_prob
        # return np.interp(times, unique_observed_times[1:], surv_func[:-1])
    else:
        for sort_idx in sort_indices:
            time = times[sort_idx]
            while leaf_time_idx < num_leaf_times:
                if unique_observed_times[leaf_time_idx] < time:
                    last_seen_surv_prob = surv_func[leaf_time_idx]
                    leaf_time_idx += 1
                else:
                    break
            output[sort_idx] = last_seen_surv_prob
        # return np.interp(times, unique_observed_times, surv_func)
    return output


class CDFRegressionKNNWeightedSurvival():
    def __init__(self, *args, **kwargs):
        """
        Arguments are the same as for `sklearn.neighbors.NearestNeighbors`.
        The simplest usage of this class is to use a single argument, which is
        `n_neighbors` for the number of nearest neighbors (Euclidean distance
        is assumed in this case). If you want to parallelize across different
        search queries, use the `n_jobs` keyword parameter (-1 to use all
        cores). To use other distances and for other details, please refer to
        the documentation for sklearn's `NearestNeighbors` class.

        *Important:* The prediction methods for this class use weighted
        k-nearest neighbors, where "k" is set equal to the `n_neighbors`
        parameter. The weights are specified through a kernel function K. In
        particular, the i-th nearest neighbor X_i for a test point x is given a
        weight of:
            K( (distance between x and X_i) / (distance between x and X_k) ).
        """
        self.NN_index_args = args
        self.NN_index_kwargs = kwargs
        self.NN_index = None

    def fit(self, X, y):
        """
        Constructs a nearest-neighbor index given training data (so that for
        a future data point, we can use the nearest-neighbor index to quickly
        find what the closest training data are to the future point).

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        y : 2D numpy array, shape = [n_samples, 2]
            Survival labels (first column is for observed times, second column
            is for event indicators). The i-th row corresponds to the i-th row
            in `X`.

        Returns
        -------
        None
        """
        self.train_y = y
        self.NN_index = NearestNeighbors(*self.NN_index_args,
                                         **self.NN_index_kwargs)
        self.NN_index.fit(X)

    def predict_surv(self, X, times, presorted_times=False,
                      limit_from_left=False, n_neighbors=None,
                      kernel_function=None):
        """
        Computes the weighted k-NN CDF estimation followed by k-NN regression
        survival probability function estimate at user-specified times.

        *Important:* The default number of nearest neighbors to use is whatever
        was specified in `args` or `kwargs` when creating an instance of this
        class (the "k" in k-NN)!

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        times : 1D numpy array
            Times to compute the survival probability function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the survival
            probability function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        n_neighbors : int, None, optional (default=None)
            Number of nearest neighbors to use. If set to None then the number
            used is whatever was passed into `args` or `kwargs` when creating
            an instance of this class.

        kernel_function : function, None, optional (default=None)
            Kernel function to use. None corresponds to unweighted k-NN
            survival analysis. If a function is specified, then the weighting
            function used is of the form
            "kernel(distance / distance to k-th nearest neighbor)".

        Returns
        -------
        output : 2D numpy array
            Survival probability function evaluated at each of the times
            specified in `times` for each feature vector. The i-th row
            corresponds to the i-th feature vector.
        """
        if kernel_function is None:
            kernel_function = lambda s: 1
        dists, indices = self.NN_index.kneighbors(X, n_neighbors=n_neighbors,
                                                  return_distance=True)
        train_y = self.train_y
        output = []
        n_times = len(times)
        for dist, idx in zip(dists, indices):
            max_dist = np.max(dist)
            weights = np.array([kernel_function(d / max_dist) for d in dist])
            zero_weight = (weights == 0)
            if zero_weight.sum() > 0:
                weights_subset = weights[~zero_weight]
                if weights_subset.size > 0:
                    labels_subset = train_y[idx[~zero_weight]]
                else:
                    output.append(np.ones(n_times))
                    continue
            else:
                labels_subset = train_y[idx]
                weights_subset = weights

            # step 1
            weighted_edf_times, weighted_edf = \
                compute_weighted_edf(labels_subset[:, 0], weights_subset)
            one_minus_weighted_edf = 1 - weighted_edf
            if weighted_edf[0] < 1 and weighted_edf_times[0] > 0:
                weighted_edf_times = \
                    np.concatenate(([0.], weighted_edf_times))
                weighted_edf = \
                    np.concatenate(([1.], weighted_edf))

            # step 2
            denoms = np.interp(labels_subset[:, 0], weighted_edf_times,
                               weighted_edf)
            neg_log_S_est = np.zeros(len(times))
            for time_idx, t in enumerate(times):
                neg_log_S_est[time_idx] = \
                    np.inner(labels_subset[:, 1]
                             * (labels_subset[:, 0] <= t) / denoms,
                             weights_subset)

            output.append(np.exp(-neg_log_S_est))
        return np.array(output)

    def predict_cum_haz(self, X, times, presorted_times=False,
                        limit_from_left=False, n_neighbors=None,
                        kernel_function=None):
        """
        Computes the weighted k-NN CDF estimation followed by k-NN regression
        cumulative hazard function estimate at user-specified times.

        *Important:* The default number of nearest neighbors to use is whatever
        was specified in `args` or `kwargs` when creating an instance of this
        class (the "k" in k-NN)!

        Parameters
        ----------
        X : 2D numpy array, shape = [n_samples, n_features]
            Feature vectors.

        times : 1D numpy array
            Times to compute the cumulative hazard function at.

        presorted_times : boolean, optional (default=False)
            Flag for whether `times` is already sorted.

        limit_from_left : boolean, optional (default=False)
            Flag for whether to output the function evaluated at a time just to
            the left, i.e., instead of outputting f(t) where f is the
            cumulative hazard function estimate, output:
                f(t-) := limit as t' approaches t from the left of f(t').

        n_neighbors : int, None, optional (default=None)
            Number of nearest neighbors to use. If set to None then the number
            used is whatever was passed into `args` or `kwargs` when creating
            an instance of this class.

        kernel_function : function, None, optional (default=None)
            Kernel function to use. None corresponds to unweighted k-NN
            survival analysis. If a function is specified, then the weighting
            function used is of the form
            "kernel(distance / distance to k-th nearest neighbor)".

        Returns
        -------
        output : 2D numpy array
            Cumulative hazard function evaluated at each of the times
            specified in `times` for each feature vector. The i-th row
            corresponds to the i-th feature vector.
        """
        if kernel_function is None:
            kernel_function = lambda s: 1
        dists, indices = self.NN_index.kneighbors(X, n_neighbors=n_neighbors,
                                                  return_distance=True)
        train_y = self.train_y
        output = []
        n_times = len(times)
        for dist, idx in zip(dists, indices):
            max_dist = np.max(dist)
            weights = np.array([kernel_function(d / max_dist) for d in dist])
            zero_weight = (weights == 0)
            if zero_weight.sum() > 0:
                weights_subset = weights[~zero_weight]
                if weights_subset.size > 0:
                    labels_subset = train_y[idx[~zero_weight]]
                else:
                    output.append(np.ones(n_times))
                    continue
            else:
                labels_subset = train_y[idx]
                weights_subset = weights

            # step 1
            weighted_edf_times, weighted_edf = \
                compute_weighted_edf(labels_subset[:, 0], weights_subset)
            one_minus_weighted_edf = 1 - weighted_edf
            if weighted_edf[0] < 1 and weighted_edf_times[0] > 0:
                weighted_edf_times = \
                    np.concatenate(([0.], weighted_edf_times))
                weighted_edf = \
                    np.concatenate(([1.], weighted_edf))

            # step 2
            denoms = np.interp(labels_subset[:, 0], weighted_edf_times,
                               weighted_edf)
            neg_log_S_est = np.zeros(len(times))
            for time_idx, t in enumerate(times):
                neg_log_S_est[time_idx] = \
                    np.inner(labels_subset[:, 1]
                             * (labels_subset[:, 0] <= t) / denoms,
                             weights_subset)

            output.append(neg_log_S_est)
        return np.array(output)


def compute_weighted_edf(obs, weights=None):
    """
    Computes a weighted empirical distribution function.

    Parameters
    ----------
    obs : 1D numpy array
        Observations to construct the weighted empirical distribution from.

    weights : 1D numpy array, None, optional (default=None)
        Nonnegative weights for the observations. The i-th weight corresponds
        to the i-th value in `obs`. None refers to using uniform weights,
        i.e., each point has weight 1/len(obs).

    Returns
    -------
    sorted_unique_obs : 1D numpy array
        Sorted unique observations in ascending order.

    weighted_edf : 1D numpy array
        The weighted empirical distribution function evaluated at each of the
        values in `sorted_unique_obs`, in the same order.
    """
    if weights is None:
        weights = np.ones(len(obs))
        weights /= weights.shape[0]

    sorted_unique_obs = np.sort(np.unique(obs))
    obs_to_idx = {obs: idx for idx, obs in enumerate(sorted_unique_obs)}
    weighted_edf = np.zeros(len(sorted_unique_obs))
    for x, w in zip(obs, weights):
        weighted_edf[obs_to_idx[x]] += w

    weighted_edf = np.cumsum(weighted_edf)
    return sorted_unique_obs, weighted_edf

def _fit_leaf(y):
    """
    Computes leaf node information given survival labels (observed times and
    event indicators).

    Parameters
    ----------
    y : 2D numpy array, shape=[n_samples, 2]
        The two columns correspond to observed times and event indicators.

    Returns
    -------
    tree : dictionary
        The leaf node information stored as a dictionary. Specifically, the
        key-value pairs of this dictionary are as follows:
        - 'times': stores the sorted unique observed times
        - 'event_counts': in the same order as `times`, the number of events
            at each unique observed time
        - 'at_risk_counts': in the same order as `times`, the number of
            subjects at risk at each unique observed time
        - 'surv': in the same order as `times`, the Kaplan-Meier survival
            probability estimate at each unique observed time
        - 'cum_haz': in the same order as `times`, the Nelson-Aalen cumulative
            hazard estimate at each unique observed time
    """
    if len(y.shape) == 1:
        y = y.reshape(1, -1)

    sorted_unique_observed_times = np.sort(np.unique(y[:, 0]))
    num_unique_observed_times = len(sorted_unique_observed_times)
    time_to_idx = {time: idx
                   for idx, time in enumerate(sorted_unique_observed_times)}
    event_counts = np.zeros(num_unique_observed_times)
    dropout_counts = np.zeros(num_unique_observed_times)
    at_risk_counts = np.zeros(num_unique_observed_times)
    at_risk_counts[0] = len(y)

    for observed_time, event_ind in y:
        idx = time_to_idx[observed_time]
        if event_ind:
            event_counts[idx] += 1
        dropout_counts[idx] += 1

    for idx in range(num_unique_observed_times - 1):
        at_risk_counts[idx + 1] = at_risk_counts[idx] - dropout_counts[idx]

    surv_prob = 1.
    cum_haz = 0.
    surv_func = np.zeros(num_unique_observed_times)
    cum_haz_func = np.zeros(num_unique_observed_times)
    for idx in range(num_unique_observed_times):
        frac = event_counts[idx] / at_risk_counts[idx]
        surv_prob *= 1 - frac
        cum_haz += frac
        surv_func[idx] = surv_prob
        cum_haz_func[idx] = cum_haz

    return {'times': sorted_unique_observed_times,
            'event_counts': event_counts,
            'at_risk_counts': at_risk_counts,
            'surv': surv_func,
            'cum_haz': cum_haz_func}



#  This function tests the distance function in the distance.py file

import unittest
import proxsurv.cbrsurv.distance_functions as distance_functions
import numpy as np

class TestDistance(unittest.TestCase):
    def test_distance(self):
        self.assertEqual(distance_functions.distance_trapz(np.array([2,2,4]),np.array([1,3,1]),np.array([6.7,8.2,9.0])),1.3478255009454345)
        self.assertEqual(distance_functions.distance_max(np.array([2,2,4]),np.array([1,3,1]),np.array([6.7,8.2,9.0])),3)
        self.assertEqual(distance_functions.distance_euclidean(np.array([2,2,4]),np.array([1,3,1]),np.array([6.7,8.2,9.0])),3.3166247903554)
        self.assertEqual(distance_functions.distance_euler(np.array([2,2,4]),np.array([1,3,1]),np.array([6.7,8.2,9.0])),4.608693648394067)




if __name__ == '__main__':
    unittest.main()
